#!/usr/bin/env python3

"""edgelist_anti.csv or edgelist_cyclo.csv must be in the current
directory."""

import networkx as nx
import csv
import shapefile
from os import path
from networkx import nx_agraph
import os

def read_eddy_graph(edgelist, shp_tr_dir = None, read_interp = True):
    """shp_tr_dir is for visible eddies."""

    G = nx.DiGraph()

    if os.access(edgelist, os.R_OK):
        with open(edgelist) as f:
            reader = csv.reader(f, delimiter = " ", skipinitialspace = True)

            # Skip title lines:
            next(reader) 
            next(reader)

            for row in reader:
                k1, i1, k2, i2 = (int(row[0]), int(row[1]), int(row[2]),
                                  int(row[3]))

                try:
                    weight = float(row[4])
                except IndexError:
                    weight = None

                G.add_edge((k1, i1), (k2, i2), weight = weight)

        if read_interp:
            # Try to read and set attributes of interpolated eddies:

            # Assuming that the directory containing the interpolated
            # eddies is in the same location as edgelist:
            dir_edgelist = path.dirname(edgelist)
            extr_file = path.join(dir_edgelist, "SHP_triplet", "extremum.shp")

            if os.access(extr_file, os.R_OK):
                set_attribute(G, extr_file)
            else:
                print("Trying to read data on interpolated eddies but ",
                      f"{extr_file} is not readable")

        if shp_tr_dir is not None:
            # Read and set attributes of Visible eddies:
            extr_file = path.join(shp_tr_dir, "extremum")
            set_attribute(G, extr_file)
    else:
        print(edgelist, "not readable")

    return G

def set_attribute(G, extr_file):
    with shapefile.Reader(extr_file) as s_read:
        for shape_rec in s_read:
            n = (shape_rec.record.days_1950, shape_rec.record.eddy_index)
            
            if (shape_rec.record.valid == 1
                or shape_rec.record.interpolat == 1) and n in G:
                G.add_node(n,
                           coordinates = tuple(shape_rec.shape.points[0]),
                           interpolat = shape_rec.record.interpolat == 1,
                           speed = shape_rec.record.speed,
                           ssh = shape_rec.record.ssh)
    
def to_eddy_agraph(G):
    my_subgraphs = {}

    for n in G:
        if n[0] not in my_subgraphs: my_subgraphs[n[0]] = []
        my_subgraphs[n[0]].append(n)

    A = nx_agraph.to_agraph(G)

    for d, s in my_subgraphs.items():
        A.add_subgraph(s, rank = "same")

    for n in G:
        if "interpolat" in G.nodes[n] and G.nodes[n]["interpolat"]:
            A_node = A.get_node(n)
            A_node.attr["shape"] = "box"

    return A

def add_component(G, source):
    """Add the weakly-connected component containing node source.

    The component is a set of nodes. Note this is not the same as the
    union of source, its descendants and ancestors.

    """

    if "component" in G.nodes[source]:
        print(f"add_component: component containing {source} already added")
    else:
        undirected_G = nx.to_undirected(G)
        node_generator = nx.dfs_preorder_nodes(undirected_G, source)
        component = set(node_generator)

        if "component_list" in G.graph:
            G.graph["component_list"].append(component)
        else:
            G.graph["component_list"] = [component]

        for n in component: G.nodes[n]["component"] = component

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--gv", help = "Create Graphviz file",
                        action = "store_true")
    parser.add_argument("-n", "--node",
                        help = "Select component containing node", nargs = 2,
                        type = int, metavar = ("days_1950", "eddy_index"))
    parser.add_argument("-m", "--merging",
                        help = "Report merging and splitting events",
                        action = "store_true")
    parser.add_argument("--interp", help = "Read interpolated eddies",
                        action = "store_true")
    parser.add_argument("orientation", help ="Choose a graph",
                        choices = ["anti", "cyclo"])
    args = parser.parse_args()
    
    G = read_eddy_graph(f"edgelist_{args.orientation}.csv",
                        read_interp = args.interp)

    if args.node is None:
        H = G
    else:
        source = tuple(args.node)
        add_component(G, source)
        H = G.subgraph(G.nodes[source]["component"])

    len_H = len(H)
    print("Number of nodes:", len_H)
    print("(This should be the number of valid eddies (outermost contour found "
          "with sufficient area), plus the number of interpolated eddies, "
          "minus the number of isolated nodes.)")
    print("Number of edges:", H.number_of_edges())
    print("Number of weakly connected components:",
          nx.number_weakly_connected_components(H))
    n1 = len_H
    n2 = len_H

    if args.merging:
        print("\nMerging:")

    for n, d in H.in_degree():
        if d >= 1:
            n2 -= 1
            if args.merging and d >= 2: print(list(H.predecessors(n)), "->", n)

    if args.merging:
        print("---")
        print("Splitting:")

    for n, d in H.out_degree():
        if d >= 1:
            n1 -= 1
            if args.merging and d >= 2: print(n, "->", list(H.successors(n)))

    if args.merging:
        print("---")

    print("Number of leaves:", n1)
    print("Number of roots:", n2)
    
    if args.gv:
        A = to_eddy_agraph(H)
        A.write(f"eddy_graph_{args.orientation}.gv")
        print(f'Created file "eddy_graph_{args.orientation}.gv".')
